'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _mysql = require('mysql');

var _mysql2 = _interopRequireDefault(_mysql);

var _path = require('path');

var _path2 = _interopRequireDefault(_path);

var _fs = require('fs');

var _fs2 = _interopRequireDefault(_fs);

var _utils = require('./utils');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _objectWithoutProperties(obj, keys) { var target = {}; for (var i in obj) { if (keys.indexOf(i) >= 0) continue; if (!Object.prototype.hasOwnProperty.call(obj, i)) continue; target[i] = obj[i]; } return target; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var defaultConnectionOptions = {
  password: '',
  sqlPath: './sql',
  retryLimit: Infinity,
  connectionLimit: null, // If this is set to a numeric value, the connection will be "pooled"
  transforms: {
    undefined: 'NULL',
    '': 'NULL',
    'NOW()': 'NOW()',
    'CURTIME()': 'CURTIME()'
  }
};

var MySql = function () {

  /****************************************
    Setup and Initialization
  *****************************************/

  function MySql(options) {
    _classCallCheck(this, MySql);

    options = _extends({}, defaultConnectionOptions, options);

    var _options = options,
        sqlPath = _options.sqlPath,
        retryLimit = _options.retryLimit,
        transforms = _options.transforms,
        connectionOptions = _objectWithoutProperties(_options, ['sqlPath', 'retryLimit', 'transforms']);

    this.connectionTries = 0;
    this.retryLimit = retryLimit;

    // Events that can be hooked into
    this.events = {
      connectionAttempt: null,
      connectionSuccess: null,
      connectionError: null,
      connectionLost: null,
      connectionTriesLimitReached: null,
      sqlError: null
    };

    // Misc Setup
    this.settings = { sqlPath: sqlPath, transforms: transforms };
    this.middleware = {
      onBeforeQuery: [],
      onResults: []
    };

    // Connect
    var isPool = Number.isInteger(connectionOptions.connectionLimit) && connectionOptions.connectionLimit > 0;
    this.connection = this.connect(connectionOptions, isPool);

    // Finish setting up pooling
    if (isPool) this.initPool(connectionOptions);

    // Setup Transactions (after connection)
    this.initTransactions();
  }

  /**
   * Events
   */

  _createClass(MySql, [{
    key: 'on',
    value: function on(eventName, cb) {
      if (typeof eventName !== 'string') throw new Error('You must pass a string value for MySQL Chassis .on() events');
      if (!this.events.hasOwnProperty(eventName)) throw new Error('MySQL Chassis does not have a db.on(\'' + eventName + '\') event. If you\'re trying to access the underlying connection events from MySQLJS, use db.connection.on instead.');
      this.events[eventName] = cb;
    }
  }, {
    key: 'emit',
    value: function emit(eventName) {
      var _events;

      if (typeof eventName !== 'string') throw new Error('You must pass a string value for MySQL Chassis .on() events');
      if (!this.events.hasOwnProperty(eventName)) throw new Error('MySQL Chassis does not have a db.on(\'' + eventName + '\') event. If you\'re trying to access the underlying connection events from MySQLJS, use db.connection.on instead.');
      if (typeof this.events[eventName] !== 'function') return;

      for (var _len = arguments.length, args = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
        args[_key - 1] = arguments[_key];
      }

      (_events = this.events)[eventName].apply(_events, args);
    }

    /**
     * MySQL Transactions
     */

  }, {
    key: 'initTransactions',
    value: function initTransactions() {
      var _this = this;

      this.transactions = {};
      this.transactions.begin = (0, _utils.promisify)(function (options, cb) {
        return _this.connection.beginTransaction(options, cb);
      });
      this.transactions.commit = (0, _utils.promisify)(function (options, cb) {
        return _this.connection.commit(options, cb);
      });
      this.transactions.rollback = (0, _utils.promisify)(function (options, cb) {
        return _this.connection.rollback(options, cb);
      });
    }

    /**
     * Connect to the database using the underlying MySQLJS lib
     * This method attempts to reconnect when the connection is lost
     */

  }, {
    key: 'connect',
    value: function connect(connectionOptions) {
      var _this2 = this;

      var pool = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;

      this.connectionTries++;
      if (this.connectionTries > this.retryLimit) {
        this.emit('connectionTriesLimitReached', this.connectionTries - 1);
        return;
      }
      this.emit('connectionAttempt', this.connectionTries);
      if (connectionOptions._connection) return connectionOptions._connection;

      // Connection
      var connection = pool ? _mysql2.default.createPool(connectionOptions) : _mysql2.default.createConnection(connectionOptions);

      // Test initial connection
      connection[pool ? 'getConnection' : 'connect'](function (err) {
        // This error occurs when we cannot make a connection
        if (err) {
          _this2.emit('connectionError', err);
          setTimeout(function () {
            _this2.connect(connectionOptions);
          }, 1000);
        } else {
          _this2.emit('connectionSuccess', _this2.connectionTries);
          _this2.connectionTries = 0;
        }
      });

      // This event fires when an established connection is lost
      connection.on('error', function (err) {
        _this2.emit('connectionLost', err);
        if (err.code === 'PROTOCOL_CONNECTION_LOST') {
          _this2.connect(connectionOptions);
        } else {
          throw err;
        }
      });

      return connection;
    }

    /**
     *  Initialize Pooling Methods
     */

  }, {
    key: 'initPool',
    value: function initPool(connectionOptions) {
      var _this3 = this;

      this.getConnection = function (cb) {
        return _this3.connection.getConnection(function (err, connection) {
          if (err) return cb(err);
          var chassisified = connection && new MySql(_extends({}, connectionOptions, {
            _connection: connection
          }));

          if (!chassisified) return cb(new Error('Could not return new connection'));

          return cb(null, chassisified);
        });
      };

      if (this.connection.release) {
        // has no callback, no need to be promisified
        this.release = this.connection.release.bind(this.connection);
      }
    }

    /****************************************
      SQL
    *****************************************/

    /**
     * Run a SELECT statement
     */

  }, {
    key: 'select',
    value: function select(sql) {
      var values = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

      return this.query(sql, values).then(function (results) {
        return results.rows;
      });
    }

    /**
     * Run a SELECT statement from a file
     */

  }, {
    key: 'selectFile',
    value: function selectFile(filename) {
      var values = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

      return this.queryFile(filename, values).then(function (results) {
        return results.rows;
      });
    }

    /**
     * Build and run a simple SELECT statement from arguments
     */

  }, {
    key: 'selectWhere',
    value: function selectWhere(fields, table, where) {
      where = this.sqlWhere(where);
      if (typeof table !== 'string') throw new Error('selectWhere table argument must be a string');
      if (typeof fields === 'string') fields = fields.split(',');
      if (Array.isArray(fields)) fields = fields.map(function (field) {
        return '`' + field.trim() + '`';
      }).join(', ');
      return this.select('SELECT ' + fields + ' FROM `' + table + '` ' + where);
    }

    /**
     * Build and run an INSERT statement
     */

  }, {
    key: 'insert',
    value: function insert(table) {
      var values = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

      var sql = 'INSERT INTO `' + table + '` SET ' + this.createInsertValues(values);
      return this.query(sql);
    }

    /**
     * Build and run an UPDATE statement
     */

  }, {
    key: 'update',
    value: function update(table, values, where) {
      var sql = 'UPDATE `' + table + '` SET ' + this.createInsertValues(values) + ' ' + this.sqlWhere(where);
      return this.query(sql);
    }

    /**
     * Try running an INSERT statement. If the record already exists, run an update statement.
     * This takes advantage of MySQL's `ON DUPLICATE KEY UPDATE` feature.
     */

  }, {
    key: 'insertUpdate',
    value: function insertUpdate(table) {
      var values = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

      values = this.createInsertValues(values);
      var sql = 'INSERT INTO `' + table + '` SET ' + values + ' ON DUPLICATE KEY UPDATE ' + values;
      return this.query(sql);
    }

    /**
     * Try running an INSERT statement. If the record already exists, ingore the request without error.
     * This takes advantage of MySQL's `INSERT IGNORE` feature.
     */

  }, {
    key: 'insertIgnore',
    value: function insertIgnore(table, values) {
      var sql = 'INSERT IGNORE INTO `' + table + '` SET ' + this.createInsertValues(values);
      return this.query(sql);
    }

    /**
     * Insert multiple values in one statement. This method creates an sql statement similar to:
     *  INSERT INTO table
     *    (a,b,c)
     *  VALUES
     *    (1,2,3),
     *    (4,5,6),
     *    (7,8,9);
     */

  }, {
    key: 'insertMultiple',
    value: function insertMultiple(table, objects, cols) {
      var _this4 = this;

      // If only two arguments are passed, the second argument becomes objects
      // and we will derive `cols` from the first object
      if (cols === undefined) cols = Object.keys(objects[0]);

      var values = objects.map(function (obj) {
        var uniformObj = [];
        obj = _this4.transformValues(obj);
        cols.forEach(function (col) {
          uniformObj.push(obj[col] || 'null');
        });
        return '(' + uniformObj.join(',') + ')';
      });

      var sql = 'INSERT INTO `' + table + '` (`' + cols.join('`,`') + '`) VALUES ' + values.join(',');
      return this.query(sql);
    }

    /**
     * Build and run a DELETE statement
     */

  }, {
    key: 'delete',
    value: function _delete(table, where) {
      var sql = 'DELETE FROM `' + table + '` ' + this.sqlWhere(where);
      return this.query(sql);
    }

    /**
     * Prepare and run a query with bound values. Return a promise
     */

  }, {
    key: 'query',
    value: function query(originalSql) {
      var _this5 = this;

      var values = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

      return new Promise(function (resolve, reject) {

        // Apply Middleware
        var finalSql = _this5.applyMiddlewareOnBeforeQuery(originalSql, values);

        // Bind dynamic values to SQL
        finalSql = _this5.queryBindValues(finalSql, values).trim();

        _this5.connection.query({
          sql: finalSql,
          typeCast: function typeCast(field, next) {
            return field.type === 'JSON' ? JSON.parse(field.string()) : next();
          }
        }, function (err, results, fields) {
          if (err) {
            // assign `sql` and preserve `message` prop
            err.sql = finalSql;
            _this5.emit('sqlError', err);
            reject(err);
          } else {

            // Apply Middleware
            results = _this5.applyMiddlewareOnResults(originalSql, results);

            // When calling `connection.query`, the results returned are either "rows"
            // in the case of a SELECT statement, or meta results in the case of non-SELECT

            // If sql is SELECT
            if (_this5.isSelect(finalSql)) {

              // Results is the rows
              resolve({ rows: results, fields: fields, sql: finalSql });
            } else {
              resolve(_extends({}, results, { sql: finalSql }));
            }
          }
        });
      });
    }
  }, {
    key: 'queryFile',
    value: function queryFile(filename) {
      var _this6 = this;

      var values = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

      return this.getFile(filename).then(function (sql) {
        return _this6.query(sql, values);
      });
    }

    /****************************************
      SQL Helper Functions
    *****************************************/

    /**
     * Get File
     */

  }, {
    key: 'getFile',
    value: function getFile(filename) {

      // Get full path
      var filePath = _path2.default.resolve(_path2.default.join(this.settings.sqlPath, filename + (_path2.default.extname(filename) === '.sql' ? '' : '.sql')));

      return new Promise(function (resolve, reject) {
        _fs2.default.readFile(filePath, 'utf8', function (err, sql) {
          if (err) {
            reject('Cannot find: ' + err.path);
          } else {
            resolve(sql.trim());
          }
        });
      });
    }

    /**
     * Turns 'foo', into
     *       '\'foo\''
     */

  }, {
    key: 'escape',
    value: function escape(content) {
      return _mysql2.default.escape(content);
    }

    /**
     * Turns `SELECT * FROM user WHERE user_id = :user_id`, into
     *       `SELECT * FROM user WHERE user_id = 1`
     */

  }, {
    key: 'queryBindValues',
    value: function queryBindValues(query, values) {
      var _this7 = this;

      if (!values) return query;

      return query.replace(/\:(\w+)/gm, function (txt, key) {
        return values.hasOwnProperty(key) ? _this7.escape(values[key]) : txt;
      });
    }

    /**
     * Turns {user_id: 1, age: null}, into
     *       "WHERE user_id = 1 AND age IS NULL"
     */

  }, {
    key: 'sqlWhere',
    value: function sqlWhere(where) {
      if (!where) return '';
      if (typeof where === 'string') return where;

      var whereArray = [];

      for (var key in where) {
        var value = where[key];
        if (value === null) {
          whereArray.push('`' + key + '` IS NULL');
        } else {
          whereArray.push('`' + key + '` = ' + this.escape(value));
        }
      }

      return 'WHERE ' + whereArray.join(' AND ');
    }

    /**
     * Turns {first_name: 'Brad', last_name: 'Westfall'}, into
     *       `first_name` = 'Brad', `last_name` = 'Westfall'
     */

  }, {
    key: 'createInsertValues',
    value: function createInsertValues(values) {
      var valuesArray = [];
      var transformedValues = this.transformValues(values);

      for (var key in transformedValues) {
        var value = transformedValues[key];
        valuesArray.push('`' + key + '` = ' + value);
      }

      return valuesArray.join();
    }

    /**
     * If the argument values match the keys of the this.transforms
     * object, then use the transforms value instead of the supplied value
     */

  }, {
    key: 'transformValues',
    value: function transformValues(values) {
      var newObj = {};

      for (var key in values) {
        var rawValue = values[key];
        var transform = this.settings.transforms[rawValue];
        var value = this.settings.transforms.hasOwnProperty(rawValue) ? typeof transform === 'function' ? transform(rawValue, values) : transform : this.escape(rawValue);

        newObj[key] = value;
      }

      return newObj;
    }
  }, {
    key: 'isSelect',
    value: function isSelect(sql) {
      return sql.trim().toUpperCase().match(/^SELECT/);
    }

    /****************************************
      Middleware
    *****************************************/

  }, {
    key: 'onResults',
    value: function onResults(middleware) {
      if (typeof middleware !== 'function') return;
      this.middleware.onResults.push(middleware);
    }
  }, {
    key: 'onBeforeQuery',
    value: function onBeforeQuery(middleware) {
      if (typeof middleware !== 'function') return;
      this.middleware.onBeforeQuery.push(middleware);
    }
  }, {
    key: 'applyMiddlewareOnResults',
    value: function applyMiddlewareOnResults(sql, results) {
      this.middleware.onResults.forEach(function (middleware) {
        results = middleware(sql, results);
      });
      return results;
    }
  }, {
    key: 'applyMiddlewareOnBeforeQuery',
    value: function applyMiddlewareOnBeforeQuery(sql, values) {
      this.middleware.onBeforeQuery.forEach(function (middleware) {
        sql = middleware(sql, values);
      });
      return sql;
    }
  }]);

  return MySql;
}();

exports.default = MySql;